#pragma once

/*
 Name:     ArduinoSoftI2c-Test
 Author:   Ulli (https://ullisroboterseite.de)
 Docs:     https://ullisroboterseite.de/arduino-soft-i2c.html
 See:      https://extremeelectronics.co.in/avr-tutorials/software-i2c-library-for-avr-mcus/

 */

 // Version 1.0 (2022-11-03)
 // -------------------------
 // - Initial version

 // Version 1.1 (2022-11-04)
 // -------------------------
 // - Error reading multiple bytes fixed

 // Version 1.2 (2022-11-07)
 // -------------------------
 // - writePacket: Parameter sendStop added
 // - readFromReagister: Parameter repeatedStart added

 // Version 1.3 (2022-11-08)
 // -------------------------
 // - begin returns true on success


 // Version 1.4 (2022-11-13)
 // -------------------------
 // - some Stop conditions had been missing
 // 

 // Version 1.5 (2022-11-14)
 // -------------------------
 // - methods return an error code. Error codes are negative integers.
 // - timeout for SCL HIGH i.e. cannot control the SCL line.
 // 

#include <Arduino.h>

class UrsArduinoSoftI2cExClass {
public:
   enum ErrorCode {
      InvalidSda = -1, // Invalid number for SDA pin.
      InvalidScl = -2, // Invalid number for SCL pin.
      SdaError = -3,   // Cannot control SDA line (cannot set SDA to HIGH level)
      SclError = -4,   // Cannot control SCL line (cannot set SCL to HIGH level, timeout)
      NoSlave = -5     // No slave responded to the device address 
   };


   // Numer of milliseconds for SCL to go HIGH.
   // Prevents infinite loops if the slave does not respond (cannot set SCL to HIGH)
   uint8_t timeout = 10;

   // Initializes the Soft I2C Engine. Must be called before using any other lib functions.
   // sda: Pin number of SDA pin.
   // scl: Pin number of SCL pin.
   // returns 0: Success or error code InvalidSda, InvalidScl, SdaError, SclError
   int8_t begin(uint8_t sda, uint8_t scl);

   // Sends data to an I2C slave.
   // deviceAdr: Address of the device [0..127].
   // pData: Pointer to the data to be send.
   // len: Number of bytes to be sent.
   // sendStop: if true, a stop consition is send after the byte
   // returns Number of bytes sent or error codes SclError (timeout) or NoSlave
   int writePacket(uint8_t deviceAdr, uint8_t* pData, uint8_t len, bool sendStop = true);

   // Reads data from an I2C slave. 
   // deviceAdr: Address of the device [0..127].
   // pData: Pointer to the buffer for storing the data.
   // len: Number of bytes to be read.
   // returns Number of bytes read, either 0 or len. Or error codes SclError (timeout) or NoSlave
   int readPacket(uint8_t deviceAdr, uint8_t* pData, uint8_t len);

   // Sends data to an I2C slave at register index.
   // deviceAdr: Address of the device [0..127].
   // regIndex: Index to register position to write to.
   // pData: Pointer to the data to be send.
   // len: Number of bytes to be sent.
   // returns Number of bytes sent or error codes SclError (timeout) or NoSlave
   int writeToRegister(uint8_t deviceAdr, uint8_t regIndex, uint8_t* pData, uint8_t len);

   // Reads data from an I2C slave. 
   // deviceAdr: Address of the device [0..127].
   // regIndex: Index to register position to read frrom.
   // pData: Pointer to the buffer for storing the data.
   // len: Number of bytes to be read.
   // repeatedStart: if true a Repeated Start (instead of Stop) is send between write register index and read data.
   // returns Number of bytes read, either 0 or len. Or error codes SclError (timeout) or NoSlave
   int readFromRegister(uint8_t deviceAdr, uint8_t regIndex, uint8_t* pData, uint8_t len, bool repeatedStart = true);

protected:
   const uint16_t delayLoops = 5; // Gives a period of 9.5us on a 16MHz Arduino Uno = 105 kHz
   volatile uint8_t* sdaDDR; // Pointer to the DDR register of the port with the SDA pin.
   volatile uint8_t* sdaPIN; // Pointer to the PIN register of the port with the SDA pin.
   volatile uint8_t* sclDDR; // Pointer to the DDR register of the port with the SCL pin.
   volatile uint8_t* sclPIN; // Pointer to the PORT register of the port with the SCL pin.

   uint8_t sdaMask, sclMask; // Bit-Masks for the pins.

   // Sets SDA pin to Output, bus line goes low.
   inline void sdaLow() {
      *sdaDDR |= sdaMask;
      _delay_loop_2(delayLoops);
   }

   // Sets SDA pin to input, bus line goes HIGH or is controlled by the slave.
   inline void sdaHigh() {
      *sdaDDR &= ~sdaMask;
      _delay_loop_2(delayLoops);
   }

   // Sets SCL pin to Output, bus line goes low.
   inline void sclLow() {
      *sclDDR |= sclMask;
      _delay_loop_2(delayLoops);
   }

   // Returns the SDA level
   inline uint8_t getSda() {
      return (*sdaPIN & sdaMask);
   }

   // Returns the SCL level
   inline uint8_t getScl() {
      return (*sclPIN & sclMask);
   }

   // Sets SCL pin to input, bus line goes HIGH or is controlled by the slave.
   inline int8_t sclHigh() {
      *sclDDR &= ~sclMask;
      _delay_loop_2(delayLoops);
      auto end = millis() + timeout;
      while (!getScl() && (millis() < end)); // Able to control SCL?
      if (!getScl())
         return SclError;
      return 0;
   }

   int8_t Start(); // Generates a START condition on the bus. Returns 0: success or SclError: timeout

   // Generates a STOP condition on the bus. Returns 0: success or SclError: timeout
   int8_t Stop();

   // Sends a byte to the slave. 
   // Returns zero if the slave acknowledges receipt of data, 1 or error code SclError otherwise.
   int8_t Write(uint8_t data);

   // Sends the device address (a byte) to the slave. 
   // Returns zero if the slave acknowledges receipt of data, 1 or error codes SclError or NoSlave otherwise.
   int8_t WriteAddress(uint8_t data) {
      auto rtc = Write(data);
      if (rtc == 1)
         return NoSlave;
      return rtc;
   }

   // Reads a byte from slave.
   // ack: true, if more bytes are to be read, otherwise false for the last byte
   // returns positive value = byte read or error code SclError
   int Read(bool ack);
};

extern UrsArduinoSoftI2cExClass SoftI2cEx;
